package com.hisschemoller.view.components.audio 
{
	import com.hisschemoller.model.vo.AudioEventVO;
	import com.hisschemoller.model.vo.FLARResultVO;
	import com.hisschemoller.utils.midi.MidiUtils;	
	/**
	 * @author Wouter Hisschemöller
	 * (c) Aug 23, 2009
	 */
	public class AudioEngine 
	{
		private var mDriver : AudioDriver;
		private var mGenerators : Vector.<AudioEventVO> = new Vector.<AudioEventVO>();
		
		private var _pentatonicIntervalsUp : Vector.<int> = Vector.<int>([0, 3, 5, 6, 7, 10, 12, 15, 17, 18, 19, 22, 24]);
		private var _pentatonicIntervalsDown : Vector.<int> = Vector.<int>([0, -2, -5, -6, -7, -9, -12, -14, -17, -18, -19, -21, -24]);
		private var _lastNote : int = 1000;
		private var _timeInSamples : int = 0;
		private var _bpm : int = 125;
		private var _measureInSamples : int;
		private var _32thInSamples : int;
		private static const ARP_LENGTH : int = 32;
		private var _nextArpPosition : int;

		public function AudioEngine()
		{
			_measureInSamples = AudioDriver.SAMPLE_RATE * (240 / _bpm);
			_32thInSamples = AudioDriver.SAMPLE_RATE * (7.5 / _bpm);
			
			mDriver = new AudioDriver(this);
		}

		/**
		 * 
		 */
		public function update(result : FLARResultVO) : void
		{
			if(result)
			{
				/** Get note to play from vertical position . */
				
				var scaledY : int = Math.round(result.relativeY * -16);
				scaledY = Math.max((_pentatonicIntervalsDown.length -1) * -1, Math.min(scaledY, _pentatonicIntervalsUp.length - 1));
				
				if(scaledY >= 0) var note : int = _pentatonicIntervalsUp[scaledY];
				else note = _pentatonicIntervalsDown[Math.abs(scaledY)];
				
				if(note == _lastNote)
				{
					var noteChanged : Boolean = false;
				}
				else
				{
					noteChanged = true;
					var frequency : Number = MidiUtils.pitchToFrequency(72 + note);
					// log("note: " + note + ", frequency: " + frequency);
				}
				
				_lastNote = note;
				
				/** Get vibrato from horizontal position. */
				
				if(result.relativeX < 0)
				{
					var vibrato : Number = result.relativeX * -1; /** 0 ... 1 */
					var vibratoChanged : Boolean = true;
				}
				else
				{
					vibrato = 0;
					vibratoChanged = false;
				}
				
				/** Get arpeggio from horizontal position. */
				
				if(result.relativeX > 0.2)
				{
					playArpeggio(result.relativeX, 72 + note);
					return;
				}
				
				/** Create or update the generator. */  
				
				if(mGenerators.length == 0)
				{
					if(!noteChanged) return;
					
					/** If there's no generator add it. */
					var audioEventVO : AudioEventVO = new AudioEventVO();
					audioEventVO.audioGenerator = new SineGenerator(0, 0.5, frequency, vibrato);
					
					mGenerators.splice(0, 0, audioEventVO);
				}
				else
				{
					/** If there is a generator update it. */
					var generator : SineGenerator = SineGenerator(mGenerators[0].audioGenerator);
					if(noteChanged) generator.updateBaseFrequency(frequency);
					if(vibratoChanged) generator.updateVibrato(vibrato);
				}
			}
			else
			{
				/** If there is a generator stop it. */
				if(mGenerators.length > 0)
				{
					if(!SineGenerator(mGenerators[0].audioGenerator).isRelease)
					{
						log("Stop");
						_lastNote = 1000;
						mGenerators[0].audioGenerator.stop(); // removeGenerator(mGenerators[0]);
					}
				}
			}
		}
		
		/**
		 * 
		 */
		private function playArpeggio(xPosition : Number, basePitch : int) : void
		{
			/** Set the next position in arpeggio Vector. */
			_nextArpPosition = int((_timeInSamples / _measureInSamples) * ARP_LENGTH);
			
			var arpPositionInSamples : int = _nextArpPosition * _32thInSamples;
			
			while(arpPositionInSamples < _timeInSamples + AudioDriver.BUFFER_SIZE)
			{
				var deltaTime : int = arpPositionInSamples - _timeInSamples;
				
				if(deltaTime >= 0)
				{
					/** If there are generators stop them. */
					var n : int = mGenerators.length;
					while(--n > -1)
					{
						if(!SineGenerator(mGenerators[n].audioGenerator).isRelease)
						{
							log("Stop");
							mGenerators[n].audioGenerator.stop(); // removeGenerator(mGenerators[0]);
						}
					}
					
					/** Play only 16th notes when marker is close to center. */
					if(xPosition >= 0.6 || (xPosition < 0.6 && _nextArpPosition % 2 == 0))
					{
						if(xPosition >= 0.6) var gain : Number = 0.15;
						else gain = 0.3;
						
						var frequency : Number = MidiUtils.pitchToFrequency(basePitch);
						
						var audioEventVO : AudioEventVO = new AudioEventVO();
						audioEventVO.audioGenerator = new SineGenerator(deltaTime, gain, frequency, 0);
						
						mGenerators.splice(0, 0, audioEventVO);
					}
				}
				
				arpPositionInSamples += _32thInSamples;
				
				_nextArpPosition ++;
				if(_nextArpPosition >= ARP_LENGTH) _nextArpPosition = 0;
			}
		}

		/**
		 * 
		 */
		public function process(inBuffer : Vector.<Vector.<Number>>) : void
		{
			var n : int = mGenerators.length;
			
			while(--n > -1)
			{
				/** IAudioProcessor returns true if the sound is finished. */
				if(mGenerators[n].audioGenerator.process(inBuffer))
				{
					//sendNotification(NOTE_REMOVED, mGenerators[n]);
					mGenerators.splice(n, 1);
				}
			}
			
			/** Increase position within measure. */
			_timeInSamples += AudioDriver.BUFFER_SIZE;
			/** Keep position within one measure. */
			while(_timeInSamples  > _measureInSamples)
			{
				_timeInSamples -= _measureInSamples;
			}
		}

		/**
		 * 
		 */
		public function addGenerator(inAudioEventVO : AudioEventVO) : void
		{
			// first remove possible identical events that are still playing
			removeGenerator(inAudioEventVO);
			
			// do not add silent audio events (with velocity zero, volume zero)
			// besides, velocity zero means 'note off' or end of the event.
			if(inAudioEventVO.velocity == 0) return;
			
			// create the generator
			inAudioEventVO.audioGenerator = new SineGenerator(inAudioEventVO.velocity, inAudioEventVO.pitch, 0, _timeInSamples);
			
			// add the event to the list of all events.
			mGenerators.splice(0, 0, inAudioEventVO);
			
			//sendNotification(NOTE_ADDED, inAudioEventVO);
		}

		/**
		 * 
		 */
		public function removeGenerator(inAudioEventVO : AudioEventVO) : void
		{
			var n : int = mGenerators.length;
			
			while(--n > -1)
			{
				// audio events are identified by their channel and pitch
				if((mGenerators[n].channel == inAudioEventVO.channel) && (mGenerators[n].pitch == inAudioEventVO.pitch))
				{
					//sendNotification(NOTE_REMOVED, mGenerators[n]);
					mGenerators.splice(n, 1);
				}
			}
		}

		/**
		 * 
		 */
		public function removeAllGenerators() : void
		{
			//var n : int = mGenerators.length;
			
			//while(--n > -1)
			//{
				//sendNotification(NOTE_REMOVED, mGenerators[n]);
			//}
			
			mGenerators = new Vector.<AudioEventVO>();
		}
	}
}
