package com.hisschemoller.view.components.audio 
{
	/**
	 * @author Wouter Hisschemöller
	 * (c) Jul 18, 2009
	 */
	public class SineGenerator implements IAudioProcessor 
	{
		private static const LOOP_SIZE : uint = 512;
		private static var cosineWave : Vector.<Number>;
		private static var isInitiated : Boolean;
		private var _positionInCycle : Number; /** Position in cycle from 0 to 1. */
		private var _cycleProgressPerSample : Number;
		private var _gain : Number;
		private var _envelope : Number = 1;
		private var _baseFrequency : Number;
		private var _addedFrequency : Number = 0;
		private var _vibratoFrequency : Number = 0; /** In Hertz. */
		private var _vibratoAmplitude : Number = 0; /** Amount of frequency change. */
		private var _vibratoPositionInCycle : Number = 0;
		private var _vibratoCycleProgressPerSample : Number = 0;
		private var _releaseDuration : int = 22050;
		private var _releasePosition : int;
		private var _isRelease : Boolean;
		private var _isDone : Boolean;
		private var _deltaSamples : int;

		public function SineGenerator(deltaSamples : int, gain : Number, frequency : Number, vibrato : Number)
		{
			_deltaSamples = deltaSamples;
			_gain = gain;
			
			_positionInCycle = 0.25;
			_vibratoPositionInCycle = 0.25;
			
			updateBaseFrequency(frequency);
			updateVibrato(vibrato);
			
			// the first time a SineGenerator is created the sine wavetable is constructed
			if(!SineGenerator.isInitiated) SineGenerator.initClass();
		}

		/**
		 * @param inFrequency : frequency in Hertz to play
		 */
		public function updateBaseFrequency(frequency : Number) : void
		{
			_baseFrequency = frequency;
			
			updateFrequency();
		}

		/**
		 * @param inFrequency : frequency in Hertz to play
		 */
		public function updateFrequency() : void
		{
			var samplesPerCycle : Number = AudioDriver.SAMPLE_RATE / (_baseFrequency + _addedFrequency);
			_cycleProgressPerSample = 1 / samplesPerCycle;
		}
		
		/**
		 * 
		 */
		public function updateVibrato(value : Number) : void
		{
			if(value > 0.1)
			{
				value -= 0.1;
				_vibratoFrequency =  value * value * 20;
				_vibratoAmplitude = value * value * 4;
				var samplesPerCycle : Number = AudioDriver.SAMPLE_RATE / _vibratoFrequency;
				_vibratoCycleProgressPerSample = 1 / samplesPerCycle;
			}
			else
			{
				_vibratoAmplitude = 0;
				_addedFrequency = 0;
			}
		}

		/**
		 * @param inBuffer : samples to be filled with sound data
		 */
		public function process(inBuffer : Vector.<Vector.<Number>>) : Boolean
		{
			// Elke sample opnieuw de positie in de cycle berekenen.
			// Daarvoor moet je weten hoeveel de positie opschuift per sample (getal van 0 tot 1).
			// Vermenigvuldig met wavetable lengte en rond af.
			// Zoek waarde in wavetable op die positie.
			
			// For each sample in the buffer the position in the sine wave cycle must be calculated.
			var exactPosition : Number;
			var lowerNearestValue : Number;
			var higherNearestValue : Number;
			var interpolationPosition : Number;
			var value : Number;
			var vibratoAmplitude : Number;
			
			var l : Vector.<Number> = inBuffer[0];
			var r : Vector.<Number> = inBuffer[1];
			
			var numSamples : int = AudioDriver.BUFFER_SIZE;
			
			for(var i : int = _deltaSamples;i < numSamples; ++i)
			{
				/** Vibrato */
				if(_vibratoAmplitude > 0)
				{
					exactPosition = _vibratoPositionInCycle * SineGenerator.LOOP_SIZE;
					lowerNearestValue = SineGenerator.cosineWave[uint(exactPosition)];
					higherNearestValue = SineGenerator.cosineWave[Math.ceil(exactPosition)];
					interpolationPosition = exactPosition - uint(exactPosition);
					vibratoAmplitude = (lowerNearestValue * (1 - interpolationPosition)) + (higherNearestValue * interpolationPosition);
					_addedFrequency = vibratoAmplitude * 100 * _vibratoAmplitude;
					_vibratoPositionInCycle += _vibratoCycleProgressPerSample;
					if(_vibratoPositionInCycle >= 1) _vibratoPositionInCycle -= 1;
					
					updateFrequency();
				}
				
				// linear interpolation for better quality
				exactPosition = _positionInCycle * SineGenerator.LOOP_SIZE; /** Unrounded position in wavetable. */
				lowerNearestValue = SineGenerator.cosineWave[uint(exactPosition)];
				higherNearestValue = SineGenerator.cosineWave[Math.ceil(exactPosition)];
				interpolationPosition = exactPosition - uint(exactPosition);
				value = (lowerNearestValue * (1 - interpolationPosition)) + (higherNearestValue * interpolationPosition);
				
				/** A release until the note stops. */
				if(_isRelease)
				{ 
					if(_releasePosition > 0)
					{
						_envelope = _releasePosition / _releaseDuration;
						_releasePosition --;
					}
					else
					{
						_isDone = true;
					}
				}
				
				l[i] += value * _gain * _envelope;
				r[i] += value * _gain * _envelope;
				
				_positionInCycle += _cycleProgressPerSample;
				
				if(_positionInCycle >= 1) _positionInCycle = 0;
			}
			
			_deltaSamples = 0;
			
			return _isDone;
		}
		
		/**
		 * 
		 */
		public function stop() : void
		{
			_isRelease = true;
			_releasePosition = _releaseDuration;
		}
		
		public function get isRelease() : Boolean
		{
			return _isRelease;
		}

		/**
		 * Init the basic 512 sample cosine wave which is used as a lookup table.
		 * That should use less CPU than a real sine calculation for each sample.
		 * This idea is taken from the MAX/MSP Cycle object.
		 */
		private static function initClass() : void
		{
			SineGenerator.isInitiated = true;
			SineGenerator.cosineWave = new Vector.<Number>(SineGenerator.LOOP_SIZE + 1, true);
				
			for(var i : uint = 0;i < SineGenerator.LOOP_SIZE; i++)
			{
				var position : Number = (Math.PI * 2) * (i / SineGenerator.LOOP_SIZE);
				SineGenerator.cosineWave[i] = Math.cos(position);
			}
			
			// one extra sample in case of rounding errors
			SineGenerator.cosineWave[SineGenerator.LOOP_SIZE] = SineGenerator.cosineWave[0];
		}
	}
}
